﻿using UnityEngine;
using System.Collections;
using System.Text.RegularExpressions;

[RequireComponent (typeof(NavMeshAgent))]
public class Enemy : LivingEntity
{
	public enum State
	{
		Idle,
		Chasing,
		Attacking,
	}

	State currentState;

	public ParticleSystem deathEffect;

	NavMeshAgent pathfinder;
	Transform target;
	LivingEntity targetEntity;
	Material skinMaterial;

	Color originalColour;

	float attackDistanceThreshold = .5f;
	// 1 sec.
	float timeBetweenAttacks = 1;
	float damage = 1;

	float nextAttackTime;
	float myCollisionRadius;
	float targetCollisionRadius;

	bool hasTarget;

	void Awake ()
	{
		pathfinder = GetComponent<NavMeshAgent> ();

		if (GameObject.FindGameObjectWithTag ("Player") != null) {
			hasTarget = true;

			target = GameObject.FindGameObjectWithTag ("Player").transform;
			targetEntity = target.GetComponent<LivingEntity> ();

			myCollisionRadius = GetComponent<CapsuleCollider> ().radius;
			targetCollisionRadius = target.GetComponent<CapsuleCollider> ().radius;
		}
	}

	protected override void Start ()
	{
		base.Start ();

		if (hasTarget) {
			currentState = State.Chasing;
			targetEntity.OnDeath += OnTargetDeath;

			StartCoroutine (UpdatePath ());
		}
	}

	public void SetCharacteristics (float moveSpeed, int hitsToKillPlayer, float enemyHealth, Color skinColour)
	{
		pathfinder.speed = moveSpeed;
		if (hasTarget) {
			damage = Mathf.Ceil (targetEntity.startingHealth / hitsToKillPlayer);
		}
		startingHealth = enemyHealth;

		skinMaterial = GetComponent<Renderer> ().material;
		skinMaterial.color = skinColour;
		originalColour = skinMaterial.color;
	}

	// hitDirectionはプロジェクタイルが進む方向
	public override void TakeHit (float damage, Vector3 hitPoint, Vector3 hitDirection)
	{
		if (damage >= health) {
			// "Quaternion.FromToRotation (Vector3.forward, hitDirection)"の意味。
			// 青Z軸をhitDirectionつまりプロジェクタイルの進む方向に向かせる。
			// Vector3.zeroから、Vector3.forwardで軸を立てて、その軸を回転。
			Destroy (Instantiate (deathEffect.gameObject, hitPoint, Quaternion.FromToRotation (Vector3.forward, hitDirection)) as GameObject, deathEffect.startLifetime);
		}

		base.TakeHit (damage, hitPoint, hitDirection);
	}

	void OnTargetDeath ()
	{
		currentState = State.Idle;
		hasTarget = false;
	}

	void Update ()
	{
		if (hasTarget) {
			if (Time.time > nextAttackTime) {
				// 敵とプレイヤーの距離をチェック
				// TODO math.
				float sqrDstToTarget = (target.position - transform.position).sqrMagnitude;
				if (sqrDstToTarget < Mathf.Pow (attackDistanceThreshold + myCollisionRadius + targetCollisionRadius, 2)) {
					nextAttackTime = Time.time + timeBetweenAttacks;
					StartCoroutine (Attack ());
				}
			}
		}
	}

	IEnumerator Attack ()
	{
		currentState = State.Attacking;
		pathfinder.enabled = false;

		Vector3 originalPosition = transform.position;
		//Vector3 attackPosition = target.position;
		Vector3 dirToTarget = (target.position - transform.position).normalized; // 向き
		Vector3 attackPosition = target.position - dirToTarget * (myCollisionRadius); // ターゲットから、衝突判定の半径を引く

		float attackSpeed = 3;
		float percent = 0;

		skinMaterial.color = Color.red;

		bool hasAppliedDamage = false;

		// enemy atttack animation
		while (percent <= 1) {

			// アニメーションで、最大移動したとき(=プレイヤーにぶつかった)にダメージを与える。
			if (percent >= .5f && !hasAppliedDamage) {
				hasAppliedDamage = true;
				targetEntity.TakeDamage (damage);
			}

			percent += Time.deltaTime * attackSpeed;
			// 二次関数で-1なので、XY座標で山を描く(放物線のような)補間。
			float interpolation = (-Mathf.Pow (percent, 2) + percent) * 4;
			transform.position = Vector3.Lerp (originalPosition, attackPosition, interpolation);

			yield return null;
		}

		skinMaterial.color = originalColour;
		currentState = State.Chasing;
		pathfinder.enabled = true;
	}

	IEnumerator UpdatePath ()
	{
		float refreshRate = .25f; // 1/4 second.

		while (hasTarget) {
			if (currentState == State.Chasing) {
				// TODO ここ再学習 なぜnormalized?
				Vector3 dirToTarget = (target.position - transform.position).normalized;
				//Vector3 targetPosition = new Vector3 (target.position.x, 0, target.position.z);
				// 上記の元々のコードとは異なり、衝突判定の半径を考慮したところに移動する。
				//Vector3 targetPosition = target.position - dirToTarget * (myCollisionRadius + targetCollisionRadius)
				// さらにアタックアニメーションの距離分も考慮する
				Vector3 targetPosition = target.position - dirToTarget * (myCollisionRadius + targetCollisionRadius + attackDistanceThreshold / 2);

				if (!dead) {
					pathfinder.SetDestination (targetPosition);
				}
			}

			yield return new WaitForSeconds (refreshRate);
		}
	}
}
