﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;

public class MapGenerator : MonoBehaviour
{
	public Map[] maps;
	public int mapIndex;

	// タイルのスケールは1
	public Transform tilePrefab;
	public Transform obstaclePrefab;
	public Transform mapFloor;
	// Unityエディタ上で、Navigationタブから、WalkableでBakeして、
	// 通行できるフィールドとする。
	public Transform navMeshFloor;
	public Transform navMeshMaskPrefab;
	// NavMeshFloorの最大サイズ。このナビゲーションにより敵が移動するため
	// 実際のマップサイズはこれより小さくなる。
	public Vector2 maxMapSize;

	[Range (0, 1)]
	public float outlinePercent;

	public float tileSize;
	List<Coord> allTileCoords;
	Queue<Coord> shuffledTileCoords;
	// 障害物のないタイルの座標をシャッフルしたキュー
	// 敵をスポーンするときにランダムデータとして使える。
	Queue<Coord> shuffledOpenTileCoords;
	// 現在のマップのタイル一覧。
	Transform[,] tileMap;

	Map currentMap;

	// Bugfix: comment from Unity Create a Game Series (E14. game over)
	// it can subscribe to the OnNewWave event before Spawner's Start method
	// is called.
	//void Start ()
	//{
	//	FindObjectOfType<Spawner> ().OnNewWave += OnNewWave;
	//}
	void Awake ()
	{
		FindObjectOfType<Spawner> ().OnNewWave += OnNewWave;
	}

	void OnNewWave (int waveNumber)
	{
		mapIndex = waveNumber - 1;
		GenerateMap ();
	}

	public void GenerateMap ()
	{
		currentMap = maps [mapIndex];
		tileMap = new Transform[currentMap.mapSize.x, currentMap.mapSize.y];
		System.Random prng = new System.Random (currentMap.seed);

		// 座標生成
		allTileCoords = new List<Coord> ();
		for (int x = 0; x < currentMap.mapSize.x; x++) {
			for (int y = 0; y < currentMap.mapSize.y; y++) {
				allTileCoords.Add (new Coord (x, y));
			}
		}
		shuffledTileCoords = new Queue<Coord> (Utility.ShuffleArray (allTileCoords.ToArray (), currentMap.seed));


		// マップフォルダー作成
		string holderName = "Generated Map";
		if (transform.FindChild (holderName)) {
			DestroyImmediate (transform.FindChild (holderName).gameObject);
		}

		Transform mapHolder = new GameObject (holderName).transform;
		mapHolder.parent = transform; // Mapオブジェクトを親にする。

		// Spawning tiles
		//
		// マップのZY平面タイル作成
		//
		for (int x = 0; x < currentMap.mapSize.x; x++) {
			for (int y = 0; y < currentMap.mapSize.y; y++) {
				// 0.5fはQuadの半分のサイズ
				Vector3 tilePosition = CoordToPosition (x, y);
				// 赤軸90度回転
				Transform newTile = Instantiate (tilePrefab, tilePosition, Quaternion.Euler (Vector3.right * 90)) as Transform;
				newTile.localScale = Vector3.one * (1 - outlinePercent) * tileSize;
				newTile.parent = mapHolder;
				tileMap [x, y] = newTile;
			}
		}

		// Spawning obstacles
		//
		// Obstacle(障害物)を個数分作成
		//
		bool[,] obstacleMap = new bool[(int)currentMap.mapSize.x, (int)currentMap.mapSize.y];

		int obstacleCount = (int)(currentMap.mapSize.x * currentMap.mapSize.y * currentMap.obstaclePercent);
		int currentObstacleCount = 0;
		// 障害物のないタイルを保持。
		List<Coord> allOpenCoords = new List<Coord> (allTileCoords);

		for (int i = 0; i < obstacleCount; i++) {
			// ランダムに座標を取得
			Coord randomCoord = GetRandomCoord ();
			obstacleMap [randomCoord.x, randomCoord.y] = true;
			currentObstacleCount++;
			if (randomCoord != currentMap.mapCentre && MapIsFullyAccessible (obstacleMap, currentObstacleCount)) {
				float obstacleHeight = Mathf.Lerp (currentMap.minObstacleHeight, currentMap.maxObstacleHeight, (float)prng.NextDouble ());
				Vector3 obstaclePosition = CoordToPosition (randomCoord.x, randomCoord.y);
				Transform newObstacle = Instantiate (obstaclePrefab, obstaclePosition + Vector3.up * obstacleHeight / 2, Quaternion.identity) as Transform;
				newObstacle.parent = mapHolder;
				newObstacle.localScale = new Vector3 ((1 - outlinePercent) * tileSize, obstacleHeight, (1 - outlinePercent) * tileSize);

				// 障害物のマテリアルを設定
				Renderer obstacleRenderer = newObstacle.GetComponent<Renderer> ();
				Material obstacleMaterial = new Material (obstacleRenderer.sharedMaterial);
				float colourPercent = randomCoord.y / (float)currentMap.mapSize.y; // y座標でグラデーション
				obstacleMaterial.color = Color.Lerp (currentMap.foregroundColor, currentMap.backgroundColor, colourPercent);

				obstacleRenderer.sharedMaterial = obstacleMaterial;

				// 上記までで、障害物が作成されたので、障害物なしタイルの一覧から座標を削除。
				allOpenCoords.Remove (randomCoord);
			} else {
				obstacleMap [randomCoord.x, randomCoord.y] = false;
				currentObstacleCount--;
			}
		}

		// 障害物がないタイル一覧を、シャッフルしてランダムデータとして使えるように、shuffledOpenTileCoordsに保持。
		shuffledOpenTileCoords = new Queue<Coord> (Utility.ShuffleArray (allOpenCoords.ToArray (), currentMap.seed));

		//// NavMeshObstacleの作成
		// Navmesh floorは実際のステージより大きくとっているので、その側をマスキングする必要がある。
		// https://youtu.be/vQgLdFNrCN8?t=8m26s
		// 右側に生成するNavMeshObstacleの中心ポイントを求める。
		Transform maskLeft = Instantiate (navMeshMaskPrefab, Vector3.left * (currentMap.mapSize.x + maxMapSize.x) / 4f * tileSize, Quaternion.identity) as Transform;
		maskLeft.parent = mapHolder;
		maskLeft.localScale = new Vector3 ((maxMapSize.x - currentMap.mapSize.x) / 2f, 1, currentMap.mapSize.y) * tileSize; // スケールで体積設定

		Transform maskRight = Instantiate (navMeshMaskPrefab, Vector3.right * (currentMap.mapSize.x + maxMapSize.x) / 4f * tileSize, Quaternion.identity) as Transform;
		maskRight.parent = mapHolder;
		maskRight.localScale = new Vector3 ((maxMapSize.x - currentMap.mapSize.x) / 2f, 1, currentMap.mapSize.y) * tileSize; // スケールで体積設定

		Transform maskTop = Instantiate (navMeshMaskPrefab, Vector3.forward * (currentMap.mapSize.y + maxMapSize.y) / 4f * tileSize, Quaternion.identity) as Transform;
		maskTop.parent = mapHolder;
		maskTop.localScale = new Vector3 (maxMapSize.x, 1, (maxMapSize.y - currentMap.mapSize.y) / 2f) * tileSize; // スケールで体積設定

		Transform maskBottom = Instantiate (navMeshMaskPrefab, Vector3.back * (currentMap.mapSize.y + maxMapSize.y) / 4f * tileSize, Quaternion.identity) as Transform;
		maskBottom.parent = mapHolder;
		maskBottom.localScale = new Vector3 (maxMapSize.x, 1, (maxMapSize.y - currentMap.mapSize.y) / 2f) * tileSize; // スケールで体積設定

		/////
		// ナビメッシュフロアの作成
		// x軸で90度回転なので、XZではなく、XY平面が90回転。
		navMeshFloor.localScale = new Vector3 (maxMapSize.x, maxMapSize.y) * tileSize;

		// 地面当たり判定のフロアの作成
		// x軸で90度回転なので、XZではなく、XY平面が90回転。
		mapFloor.localScale = new Vector3 (currentMap.mapSize.x * tileSize, currentMap.mapSize.y * tileSize);
	}

	// Flood fill アルゴリズムで、移動できる箇所を調べて、
	// 生成したObstacleによって、道が塞がれ到達できない場合は、falseを返す。
	bool MapIsFullyAccessible (bool[,] obstacleMap, int currentObstacleCount)
	{
		bool[,] mapFlags = new bool[obstacleMap.GetLength (0), obstacleMap.GetLength (1)];
		Queue<Coord> queue = new Queue<Coord> ();
		// 起点として、センター座標は到達できる状態にしておく
		queue.Enqueue (currentMap.mapCentre);
		mapFlags [currentMap.mapCentre.x, currentMap.mapCentre.y] = true;

		int accessibleTileCount = 1;

		while (queue.Count > 0) {
			Coord tile = queue.Dequeue ();

			// check neighbor
			for (int x = -1; x <= 1; x++) {
				for (int y = -1; y <= 1; y++) {
					int neighbourX = tile.x + x;
					int neighbourY = tile.y + y;
					if (x == 0 || y == 0) {
						// 隣接がマイナスや、マップサイズ以上になっていない場合、つまりobstacleMap内の場合、
						//
						// xは調べるところ、+は中心、
						// xxx
						// x+-
						// x--
						if (neighbourX >= 0 && neighbourX < obstacleMap.GetLength (0) && neighbourY >= 0 && neighbourY < obstacleMap.GetLength (1)) {
							// mapFlagは移動できるかのフラグ、obstacleMapは、障害物であるかのマップ
							// よって、その隣接XYが、移動出来て、障害物じゃない場合、次の起点として使えるようにQueueに追加。
							if (!mapFlags [neighbourX, neighbourY] && !obstacleMap [neighbourX, neighbourY]) {
								mapFlags [neighbourX, neighbourY] = true;
								queue.Enqueue (new Coord (neighbourX, neighbourY));
								accessibleTileCount++;
							}
						}
					}
				}
			}
		}

		int targetAccessibleTielCount = (int)(currentMap.mapSize.x * currentMap.mapSize.y - currentObstacleCount);
		return targetAccessibleTielCount == accessibleTileCount;
	}

	// 原点を中心にする
	Vector3 CoordToPosition (int x, int y)
	{
		return new Vector3 (-currentMap.mapSize.x / 2f + 0.5f + x, 0, -currentMap.mapSize.y / 2f + 0.5f + y) * tileSize;
	}

	public Transform GetTileFromPosition (Vector3 position)
	{
		int x = Mathf.RoundToInt (position.x / tileSize + (currentMap.mapSize.x - 1) / 2f);
		// XZ平面なので、position.zが正しい、position.yではない。
		int y = Mathf.RoundToInt (position.z / tileSize + (currentMap.mapSize.y - 1) / 2f);

		x = Mathf.Clamp (x, 0, tileMap.GetLength (0) - 1);
		y = Mathf.Clamp (y, 0, tileMap.GetLength (1) - 1);
		return tileMap [x, y];
	}

	public Coord GetRandomCoord ()
	{
		// キューから取り出す。=Dequeue
		Coord randomCoord = shuffledTileCoords.Dequeue ();
		// キューにれる。=Enqueue
		shuffledTileCoords.Enqueue (randomCoord);
		return randomCoord;
	}

	public Transform GetRandomOpenTile ()
	{
		// キューから取り出す。=Dequeue
		Coord randomCoord = shuffledOpenTileCoords.Dequeue ();
		// キューにれる。=Enqueue
		shuffledOpenTileCoords.Enqueue (randomCoord);
		return tileMap [randomCoord.x, randomCoord.y];
	}



	[System.Serializable]
	public struct Coord
	{
		public int x;
		public int y;

		public Coord (int _x, int _y)
		{
			x = _x;
			y = _y;
		}

		public static bool operator == (Coord c1, Coord c2)
		{
			return c1.x == c2.x && c1.y == c2.y;
		}

		public static bool operator != (Coord c1, Coord c2)
		{
			// 早速上記の==演算子を使う
			return !(c1 == c2);
		}

		public override bool Equals (object obj)
		{
			
			// If parameter is null return false
			if (obj != null) {
				
				// If parameter cannot be cast to Coord return false.
				try {
					Coord p = (Coord)obj;

					// Return true if the fields match
					return (x == p.x) && (y == p.y);
				} catch (InvalidCastException) {
					return false;
				}
			}
			// obj is null.
			return false;
		}

		/*
		public override bool Equals (object obj)
		{
			// If parameter is null return false
			if (obj == null) {
				return false;
			}
			// If parameter cannot be cast to Coord return false.
			Coord p = obj as Coord;
			if ((object)p == null) {
				return false;
			}
			// Return true if the fields match
			return (x == p.x) && (y == p.y);
		}
		*/

		public bool Equals (Coord p)
		{
			if ((object)p == null) {
				return false;
			}
			// Return true if the fields match
			return (x == p.x) && (y == p.y);
		}

		public override int GetHashCode ()
		{
			return x ^ y;
		}
	}

	// マップを保存できるようにするためのクラス。
	[System.Serializable]
	public class Map
	{
		public Coord mapSize;
		[Range (0, 1)]
		public float obstaclePercent;
		public int seed;
		public float minObstacleHeight;
		public float maxObstacleHeight;
		public Color foregroundColor;
		public Color backgroundColor;

		public Coord mapCentre {
			get {
				return new Coord (mapSize.x / 2, mapSize.y / 2);
			}
		}
	}
}
